//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2018 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package v1

import (
	"errors"
	"io"
	"net/http"
	"os"
	"path/filepath"

	"github.com/gorilla/mux"
	"gitlab.servus.at/autoradio/tank/importer"
)

//******* simple binary http upload

func (api *API) UploadFileSimple() http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		vars := mux.Vars(r)
		id, err := idFromString(vars["file-id"])
		if err != nil {
			sendWebResponse(w, http.StatusBadRequest, ErrorResponse{Error: "invalid file-id: " + err.Error()})
			return
		}
		job, err := api.importer.GetJob(vars["group-id"], id)
		if err != nil {
			sendError(w, err)
			return
		}

		done, err := job.AttachUploader(uint64(r.ContentLength), r.Body)
		if err != nil {
			sendError(w, err)
			return
		}
		result := <-done
		if result == nil {
			sendError(w, errors.New("upload failed for unknown reason"))
			return
		}
		if result.Err != nil {
			sendError(w, result)
			return
		}
		sendWebResponse(w, http.StatusOK, result)
	})
}

//******* flow.js

type FlowJSFileChunk struct {
	file      *os.File
	completed chan struct{}
}

func (ch *FlowJSFileChunk) Read(p []byte) (int, error) {
	return ch.file.Read(p)
}

func (ch *FlowJSFileChunk) Write(p []byte) (int, error) {
	return ch.file.Write(p)
}

func (ch *FlowJSFileChunk) Complete() (err error) {
	if _, err = ch.file.Seek(0, io.SeekStart); err != nil {
		return
	}
	close(ch.completed)
	return
}

func (ch *FlowJSFileChunk) Cleanup() (err error) {
	filename := ch.file.Name()
	if err = ch.file.Close(); err != nil {
		return err
	}
	return os.Remove(filename)
}

type FlowJSFile struct {
	size       uint64
	chunks     []FlowJSFileChunk
	chunksPath string
	readOffset int
	job        *importer.Job
}

func newFlowJSFile(size uint64, numChunks uint, job *importer.Job) (*FlowJSFile, error) {
	f := &FlowJSFile{size: size, job: job}
	f.chunksPath = filepath.Join(job.WorkDir, "chunks")
	f.chunks = make([]FlowJSFileChunk, numChunks)
	for i := range f.chunks {
		f.chunks[i].completed = make(chan struct{})
	}
	err := os.Mkdir(f.chunksPath, 0700)
	return f, err
}

func (f *FlowJSFile) Read(p []byte) (n int, err error) {
	for {
		if f.readOffset >= len(f.chunks) {
			// we already read all the chunks
			return 0, io.EOF
		}

		// wait for chunk to be completed
		select {
		case <-f.job.Ctx.Done():
			return 0, f.job.Ctx.Err()
		case <-f.chunks[f.readOffset].completed:
		}

		if n, err = f.chunks[f.readOffset].Read(p); err == nil || err == io.EOF {
			if n > 0 {
				err = nil
				return
			}
			// we are at the end of the chunk but haven't read anything yet...
			// .. move on to the next chunk
			f.readOffset++
			continue
		}

		// we are done with this chunk, clean it up...
		if err = f.chunks[f.readOffset].Cleanup(); err != nil {
			return
		}
		// ...and proceed to the next
		f.readOffset++
		if n > 0 {
			// if this was the last chunk the next call to Read() will return (0, io.EOF)
			// for now we will return (n, nil)
			return
		}
	}
}

func (api *API) UploadFileFlowJS() http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		// TODO: implement this
		sendWebResponse(w, http.StatusNotImplemented, ErrorResponse{"flow.js upload is not yet implemented", nil})
	})
}

func (api *API) TestFileFlowJS() http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		// TODO: implement this
		sendWebResponse(w, http.StatusNotImplemented, ErrorResponse{"flow.js testing is not yet implemented", nil})
	})
}
